@@ -36,8 +36,8 @@ git archive "${ver}" | tar xf - -C "${p}"
 cd "${p}"
 
 einfo "Building autotools ..."
-sed -i "/^AC_INIT/s:git:${ver}:" configure.ac
-sed -i "1iPV := ${ver}" Makefile
+sed -i "/^AC_INIT/s:git:${ver#v}:" configure.ac
+sed -i "1iPV := ${ver#v}" Makefile
 LC_ALL=C ${MAKE} -s autotools >/dev/null
 rm -rf autom4te.cache
 cd ..
