@@ -1,9 +1,10 @@
 /*
- * Copyright 2005-2018 Gentoo Foundation
+ * Copyright 2005-2018 Gentoo Authors
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2010 Ned Ludd        - <solar@gentoo.org>
  * Copyright 2005-2014 Mike Frysinger  - <vapier@gentoo.org>
+ * Copyright 2018-     Fabian Groffen  - <grobian@gentoo.org>
  */
 
 #ifdef APPLET_qdepends
@@ -486,6 +487,7 @@ qdepends_vdb_deep_cb(q_vdb_pkg_ctx *pkg_ctx, void *priv)
 	depend_atom *as;
 	depend_atom *ac;
 	char firstmatch = 0;
+	char *sslot;
 
 	if (!q_vdb_pkg_eat(pkg_ctx, state->depend_file, &depend, &depend_len))
 		return 0;
@@ -555,10 +557,13 @@ qdepends_vdb_deep_cb(q_vdb_pkg_ctx *pkg_ctx, void *priv)
 		if (ret != 0)
 			break;
 
-		/* find the boundaries for matched atom */
+		/* find the boundaries for matched atom, dep specifications can
+		 * include built-with-use deps using [xxx] notation, so ensure
+		 * we exclude that as part of the atom */
 		while (match.rm_so > 0 && !isspace(ptr[match.rm_so - 1]))
 			match.rm_so--;
-		while (ptr[match.rm_eo] != '\0' && !isspace(ptr[match.rm_eo]))
+		while (ptr[match.rm_eo] != '\0' && ptr[match.rm_eo] != '[' &&
+				!isspace(ptr[match.rm_eo]))
 			match.rm_eo++;
 
 		snprintf(qbuf, sizeof(qbuf), "%.*s",
@@ -566,7 +571,13 @@ qdepends_vdb_deep_cb(q_vdb_pkg_ctx *pkg_ctx, void *priv)
 					ptr + match.rm_so);
 		ac = atom_explode(qbuf);
 
+		/* drop SLOT when not present in aq so we can match atoms
+		 * regardless */
+		sslot = ac->SLOT;
+		if (aq->SLOT == NULL && ac->SLOT != NULL)
+			ac->SLOT = NULL;
 		ret = atom_compare(ac, aq);
+		ac->SLOT = sslot;
 		if (ret != EQUAL) {
 			atom_implode(ac);
 			break;
